# Generated by Django 5.0.7 on 2024-11-06 11:41

import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_goals', '0002_goalprogress_message'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='goal',
            name='goals_waiting_for_worker_idx',
        ),
        migrations.AddField(
            model_name='goal',
            name='deadline',
            field=models.DateTimeField(default=django.utils.timezone.now, help_text='Goals having deadline sooner will be pursued first.'),
        ),
        migrations.AddIndex(
            model_name='goal',
            index=models.Index(condition=models.Q(('state', 'waiting_for_worker')), fields=['deadline'], name='goals_waiting_for_worker_idx'),
        ),
    ]



from django.db import connection, connections


def notify_goal_waiting_for_worker(cursor, goal_id):
    """
    Notify that the goal is waiting for a worker to pick it up.
    """
    cursor.execute("NOTIFY goal_waiting_for_worker, %s", [str(goal_id)])


def listen_goal_waiting_for_worker():
    with connection.cursor() as cursor:
        cursor.execute("LISTEN goal_waiting_for_worker")


def notify_goal_progress(goal_id, state):
    """
    Notify that the goal has changed its state.
    """
    with connections['default'].cursor() as cursor:
        channel = get_goal_progress_channel(goal_id)
        cursor.execute(f"NOTIFY {channel}, %s", [
            state,
        ])


def listen_goal_progress(goal_id):
    """
    Listen for goal progress notifications.
    """
    with connections['default'].cursor() as cursor:
        channel = get_goal_progress_channel(goal_id)
        cursor.execute(f'LISTEN {channel}')


def get_goal_progress_channel(goal_id):
    """
    Get the channel name for goal progress notifications.
    """
    return f'goal_progress_{goal_id.hex}'


def wait():
    """
    Wait for a goal progress notification.
    """
    pg_conn = connections['default'].connection
    notification_generator = pg_conn.notifies()
    for notification in notification_generator:
        notification_generator.close()
    return notification  # pylint: disable=undefined-loop-variable



import json

from django.contrib import admin, messages
from django.db import models
from django.urls import reverse
from django.urls.exceptions import NoReverseMatch
from django.utils.html import format_html, format_html_join
from django.utils.translation import gettext as _
from django_object_actions import DjangoObjectActions, action

from .models import (
    Goal, GoalDependency, GoalProgress, block_goal, unblock_retry_goal,
)
from .pickups import GoalPickup


class GoalDependencyInline(admin.TabularInline):
    model = GoalDependency
    fk_name = 'dependent_goal'
    extra = 0
    fields = (
        'precondition_goal__id',
        'precondition_goal__state',
        'precondition_goal__handler',
        'precondition_goal__created_at',
    )
    readonly_fields = fields

    def has_add_permission(self, request, obj):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    @admin.display(description='Precondition Goal ID')
    def precondition_goal__id(self, obj):
        return format_html(
            '<a href="{}">{}</a>',
            reverse(
                'admin:django_goals_goal_change',
                args=(obj.precondition_goal.id,),
            ),
            obj.precondition_goal.id,
        )

    @admin.display(description='Precondition Goal State')
    def precondition_goal__state(self, obj):
        return obj.precondition_goal.get_state_display()

    @admin.display(description='Precondition Goal Handler')
    def precondition_goal__handler(self, obj):
        return obj.precondition_goal.handler

    @admin.display(description='Precondition Goal Created At')
    def precondition_goal__created_at(self, obj):
        return obj.precondition_goal.created_at


class GoalProgressInline(admin.TabularInline):
    model = GoalProgress
    extra = 0

    def has_add_permission(self, request, obj):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


class GoalPickupInline(admin.TabularInline):
    model = GoalPickup
    extra = 0

    def has_add_permission(self, request, obj):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Goal)
class GoalAdmin(DjangoObjectActions, admin.ModelAdmin):
    list_display = (
        'id', 'state', 'handler',
        'precondition_date',
        'waiting_for_not_achieved_count',
        'created_at',
        'progress_count',
    )
    list_filter = ('state', 'precondition_date')
    search_fields = ('id',)

    fields = (
        'id',
        'state',
        'handler',
        'instructions_pre',
        'precondition_date',
        'preconditions_mode',
        'precondition_failure_behavior',
        'waiting_for_count',
        'waiting_for_not_achieved_count',
        'waiting_for_failed_count',
        'deadline',
        'created_at',
        'related_objects',
    )
    inlines = (
        GoalDependencyInline,
        GoalProgressInline,
        GoalPickupInline,
    )
    change_actions = (
        'block',
        'unblock_retry',
    )

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    def get_queryset(self, request):
        return super().get_queryset(request).annotate(
            progress_count=models.Count('progress'),
        )

    @admin.display
    def progress_count(self, obj):
        return obj.progress_count

    @admin.display(description='Instructions')
    def instructions_pre(self, obj):
        return format_html(
            '<pre style="white-space: pre-wrap;">{}</pre>',
            json.dumps(obj.instructions, indent=2),
        )

    @admin.display(description='Related Objects')
    def related_objects(self, obj):
        rows_html = []
        for field in obj._meta._relation_tree:
            if field.model._meta.app_label == 'django_goals':
                continue
            related_objects = field.model.objects.filter(**{field.name: obj})
            for related_object in related_objects:
                try:
                    object_admin_url = reverse(
                        f'admin:{related_object._meta.app_label}_{related_object._meta.model_name}_change',
                        args=(related_object.pk,),
                    )
                except NoReverseMatch:
                    object_link = format_html(
                        '<span>{related_object}</span>',
                        related_object=related_object,
                    )
                else:
                    object_link = format_html(
                        '<a href="{object_admin_url}">{related_object}</a>',
                        object_admin_url=object_admin_url,
                        related_object=related_object,
                    )
                row_html = format_html(
                    (
                        '<tr>'
                        '<td>{related_app}</td>'
                        '<td>{related_model}</td>'
                        '<td>{related_field}</td>'
                        '<td>{related_object}</td>'
                        '</tr>'
                    ),
                    related_app=related_object._meta.app_label,
                    related_model=related_object._meta.verbose_name,
                    related_field=field.verbose_name,
                    related_object=object_link,
                )
                rows_html.append(row_html)
        return format_html_join(
            '',
            (
                '<table>'
                '<thead><tr>'
                '<th>App</th>'
                '<th>Model</th>'
                '<th>Field</th>'
                '<th>Object</th>'
                '</tr></thead>'
                '<tbody>{}</tbody>'
                '</table>'
            ),
            ((row,) for row in rows_html),
        )

    @action(label=_('Unblock / retry'), methods=['POST'], button_type='form')
    def unblock_retry(self, request, obj):
        try:
            unblock_retry_goal(obj.id)
        except ValueError as e:
            self.message_user(request, str(e), level=messages.ERROR)
        else:
            self.message_user(request, _('Goal was unblocked'))

    @action(label=_('Block'), methods=['POST'], button_type='form')
    def block(self, request, obj):
        try:
            block_goal(obj.id)
        except ValueError as e:
            self.message_user(request, str(e), level=messages.ERROR)
        else:
            self.message_user(request, _('Goal was blocked'))



import pytest
from django.core.management import call_command

from django_goals.models import (
    AllDone, Goal, GoalProgress, GoalState, RetryMeLater, schedule,
)


def pursue(goal):
    return RetryMeLater()


@pytest.mark.django_db
def test_max_transitions():
    goal = schedule(pursue)
    call_command('goals_busy_worker', max_progress_count=10)
    goal.refresh_from_db()
    progress_count = goal.progress.count()
    assert progress_count == 10


def achieve(goal):
    return AllDone()


@pytest.mark.django_db
def test_max_transitions_in_single_turn():
    for _ in range(10):
        schedule(achieve)
    call_command('goals_busy_worker', max_progress_count=5)
    assert Goal.objects.filter(state=GoalState.ACHIEVED).count() == 5
    assert GoalProgress.objects.count() == 5



import contextlib
import os
import subprocess

import pytest
from django.core.management import call_command

from .goals_threaded_worker_tests import get_current_database_url


@pytest.mark.django_db(transaction=True)
def test_no_smoke():
    with worker_subprocess():
        call_command('set_goal')


@contextlib.contextmanager
def worker_subprocess():
    with subprocess.Popen(
        ['python', 'manage.py', 'goals_threaded_worker'],
        env={
            **os.environ,
            'DATABASE_URL': get_current_database_url(),
        },
    ) as p:
        try:
            yield
        finally:
            p.terminate()
            pass



import pytest
from django.core.management import call_command

from django_goals.factories import GoalFactory
from django_goals.models import (
    GoalState, PreconditionFailureBehavior, PreconditionsMode,
)


@pytest.mark.django_db
def test_goals_fsck(goal):
    goal.precondition_goals.add(
        GoalFactory(state=GoalState.ACHIEVED),
        GoalFactory(state=GoalState.WAITING_FOR_WORKER),
        GoalFactory(state=GoalState.NOT_GOING_TO_HAPPEN_SOON),
    )
    call_command('goals_fsck')
    goal.refresh_from_db()
    assert goal.waiting_for_count == 2
    assert goal.waiting_for_not_achieved_count == 2
    assert goal.waiting_for_failed_count == 1


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{'preconditions_mode': PreconditionsMode.ANY}], indirect=True)
def test_goals_fsck_any_mode(goal):
    goal.precondition_goals.add(
        GoalFactory(state=GoalState.ACHIEVED),
        GoalFactory(state=GoalState.WAITING_FOR_WORKER),
        GoalFactory(state=GoalState.WAITING_FOR_WORKER),
        GoalFactory(state=GoalState.NOT_GOING_TO_HAPPEN_SOON),
    )
    call_command('goals_fsck')
    goal.refresh_from_db()
    assert goal.waiting_for_count == 1
    assert goal.waiting_for_not_achieved_count == 3
    assert goal.waiting_for_failed_count == 1


@pytest.mark.django_db
@pytest.mark.parametrize('goal', [{
    'precondition_failure_behavior': PreconditionFailureBehavior.PROCEED,
}], indirect=True)
def test_goals_fsck_proceed_mode(goal):
    goal.precondition_goals.add(
        GoalFactory(state=GoalState.ACHIEVED),
        GoalFactory(state=GoalState.WAITING_FOR_WORKER),
        GoalFactory(state=GoalState.NOT_GOING_TO_HAPPEN_SOON),
    )
    call_command('goals_fsck')
    goal.refresh_from_db()
    assert goal.waiting_for_count == 1
    assert goal.waiting_for_not_achieved_count == 2
    assert goal.waiting_for_failed_count == 1



# Generated by Django 5.1.3 on 2024-12-08 22:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_goals', '0004_goal_waiting_for_count'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='goal',
            index=models.Index(condition=models.Q(('state', 'waiting_for_preconditions')), fields=['waiting_for_count'], name='goals_waiting_for_precond_idx'),
        ),
    ]



import contextlib
import logging
import signal
import threading

from django.core.management.base import BaseCommand

from django_goals.busy_worker import worker


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Run the worker'

    def add_arguments(self, parser):
        parser.add_argument(
            '--max-progress-count',
            type=int,
            default=float('inf'),
            help='Exit when this many progress records are created',
        )

    def handle(self, *args, max_progress_count, **options):
        with stop_signal_handler() as stop_event:
            worker(stop_event, max_progress_count=max_progress_count)


@contextlib.contextmanager
def stop_signal_handler():
    stop_event = threading.Event()

    def handler(signum, frame):
        signal_str = signal.Signals(signum).name
        logger.info('Received signal %s, stopping', signal_str)
        stop_event.set()

    with (
        set_signal_handler(signal.SIGINT, handler),
        set_signal_handler(signal.SIGTERM, handler),
    ):
        yield stop_event


@contextlib.contextmanager
def set_signal_handler(signum, handler):
    old_handler = signal.signal(signum, handler)
    try:
        yield
    finally:
        signal.signal(signum, old_handler)



import uuid

from django.core.management.base import BaseCommand
from django.db import transaction

from django_goals.models import (
    NOT_GOING_TO_HAPPEN_SOON_STATES, Goal, GoalState,
    PreconditionFailureBehavior, PreconditionsMode,
)


class Command(BaseCommand):
    """
    Check and fix the integrity of the goals system.
    """
    def handle(self, *args, **options):
        check_fix_all()


def check_fix_all():
    goal_id = uuid.UUID(int=0)
    i = 0
    while True:
        goal_id = check_fix_goal(goal_id)
        if not goal_id:
            break

        i += 1
        if i >= 1000:
            print(i, goal_id)
            i = 0

        goal_id = uuid.UUID(int=goal_id.int + 1)


@transaction.atomic
def check_fix_goal(goal_id):
    goal = Goal.objects.filter(id__gte=goal_id).order_by('id').select_for_update(
        no_key=True,
        skip_locked=True,
    ).first()
    if not goal:
        return None

    preconditions = list(goal.precondition_goals.all().select_for_update(
        no_key=True,
    ))
    waiting_for_count = 0
    waiting_for_failed_count = 0
    for pre in preconditions:
        if pre.state != GoalState.ACHIEVED:
            waiting_for_count += 1
        if pre.state in NOT_GOING_TO_HAPPEN_SOON_STATES:
            waiting_for_failed_count += 1
    waiting_for_not_achieved_count = waiting_for_count
    if goal.precondition_failure_behavior == PreconditionFailureBehavior.PROCEED:
        waiting_for_count -= waiting_for_failed_count
    if goal.preconditions_mode == PreconditionsMode.ANY:
        waiting_for_count = min(1, waiting_for_count)

    if waiting_for_count != goal.waiting_for_count:
        print(f"Goal {goal_id} waiting_for count, DB={goal.waiting_for_count}, recalculated={waiting_for_count}")
        goal.waiting_for_count = waiting_for_count
        goal.save(update_fields=['waiting_for_count'])

    if waiting_for_not_achieved_count != goal.waiting_for_not_achieved_count:
        print(f"Goal {goal_id} waiting_for_not_achieved count, DB={goal.waiting_for_not_achieved_count}, recalculated={waiting_for_not_achieved_count}")
        goal.waiting_for_not_achieved_count = waiting_for_not_achieved_count
        goal.save(update_fields=['waiting_for_not_achieved_count'])

    if waiting_for_failed_count != goal.waiting_for_failed_count:
        print(f"Goal {goal_id} waiting_for_failed count, DB={goal.waiting_for_failed_count}, recalculated={waiting_for_failed_count}")
        goal.waiting_for_failed_count = waiting_for_failed_count
        goal.save(update_fields=['waiting_for_failed_count'])

    return goal.id



from django.db import models

from .models import (
    NOT_GOING_TO_HAPPEN_SOON_STATES, WAITING_STATES, Goal, GoalState,
)


def is_goal_completed(goal):
    return (
        goal is None or  # completed some time ago and the goal was garbage collected
        goal.state == GoalState.ACHIEVED
    )


def is_goal_processing(goal):
    return goal is not None and goal.state in WAITING_STATES


def is_goal_error(goal):
    return goal is not None and goal.state in NOT_GOING_TO_HAPPEN_SOON_STATES


class GoalRelatedMixin(models.Model):
    class Meta:
        abstract = True

    processed_goal = models.OneToOneField(
        to=Goal,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    @property
    def is_completed(self):
        return is_goal_completed(self.processed_goal)

    is_done = is_completed

    @property
    def is_processing(self):
        return is_goal_processing(self.processed_goal)

    @property
    def is_error(self):
        return is_goal_error(self.processed_goal)



import os
import subprocess
from datetime import timedelta
from urllib.parse import urlparse, urlunparse

import pytest
from django.core.management import call_command
from django.db import connection
from django.utils.timezone import now

from django_goals.models import AllDone, GoalState, schedule
from django_goals.pickups import GoalPickup


def achieve(goal):
    return AllDone()


@pytest.mark.django_db(transaction=True)
def test_achieving_goals():
    goal = schedule(achieve)
    call_command('goals_threaded_worker', threads=['2'], once=True)
    goal.refresh_from_db()
    assert goal.state == GoalState.ACHIEVED


@pytest.mark.parametrize(
    ('threads_spec', 'expected_goal_state'),
    [
        (['1:1m'], GoalState.WAITING_FOR_WORKER),
        (['1:1m', '1:2d'], GoalState.ACHIEVED),
        (['1:2d', '1:1m'], GoalState.ACHIEVED),
        (['1:0s'], GoalState.WAITING_FOR_WORKER),
    ],
)
@pytest.mark.django_db(transaction=True)
def test_deadline_horizon(threads_spec, expected_goal_state):
    """ Worker should pick up goals with deadline within the deadline horizon """
    goal = schedule(achieve, deadline=now() + timedelta(days=1))
    call_command('goals_threaded_worker', threads=threads_spec, once=True)
    goal.refresh_from_db()
    assert goal.state == expected_goal_state


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize('should_commit', [True, False])
def test_exit_goal(pytestconfig, should_commit):
    """We test (in a separate process) a goal that exits and does not commit."""
    goal = schedule(achieve if should_commit else exit_goal)
    subprocess.run(
        ['python', 'manage.py', 'goals_threaded_worker', '--once'],
        env={
            **os.environ,
            'DATABASE_URL': get_current_database_url(),
        },
        cwd=pytestconfig.rootdir,
    )
    assert GoalPickup.objects.filter(goal=goal).exists() is not should_commit


def exit_goal(goal):
    os._exit(0)  # Exit the process immediately


def get_current_database_url():
    current_database_url = os.environ.get('DATABASE_URL')
    parsed = urlparse(current_database_url)
    parsed = parsed._replace(path=connection.settings_dict['NAME'])
    return urlunparse(parsed)



import sentry_sdk


class Middleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, goal, *args, **kwargs):
        headers = (goal.instructions or {}).get('sentry', {})
        transaction = sentry_sdk.continue_trace(
            headers,
            op='queue.process',
            name=goal.handler,
        )
        with sentry_sdk.start_transaction(transaction):
            transaction.set_data('messaging.message.id', str(goal.id))
            transaction.set_data('messaging.destination.name', goal.handler)
            progress = self.get_response(goal, *args, **kwargs)
            transaction.set_data('goal.final_state', goal.state)
            if progress:
                transaction.set_data('goal.progress.success', progress.success)
                transaction.set_data('goal.progress.message', progress.message)
                transaction.set_data(
                    'goal.progress.time_taken',
                    progress.time_taken.total_seconds() if progress.time_taken else None,
                )
            return progress


class ScheduleMiddleware:
    def __init__(self, schedule):
        self.schedule = schedule

    def __call__(self, goal, *args, **kwargs):
        goal.instructions = goal.instructions or {}
        goal.instructions['sentry'] = {
            "sentry-trace": sentry_sdk.get_traceparent(),
            "baggage": sentry_sdk.get_baggage(),
        }
        with sentry_sdk.start_span(
            op='queue.schedule',
            name='Django Goals schedule',
        ) as span:
            span.set_data('messaging.message.id', str(goal.id))
            span.set_data('messaging.destination.name', goal.handler)
            return self.schedule(goal, *args, **kwargs)



import logging
import queue
import threading
import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone


logger = logging.getLogger(__name__)


class GoalPickup(models.Model):
    """
    Stores what goals were picked in a transaction-independent and work-independent way.
    We use this to detect "killer tasks".
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    goal = models.ForeignKey(to='Goal', on_delete=models.CASCADE, related_name='pickups')
    created_at = models.DateTimeField(default=timezone.now)

    class Meta:
        ordering = ('-created_at',)


class PickupMonitorThread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.event_queue = queue.Queue()

    def run(self):
        logger.info('Pickup monitor thread started')

        while True:
            try:
                event, goal_id = self.event_queue.get(timeout=1)
            except queue.Empty:
                continue
            except queue.ShutDown:
                break

            if event == 'pickup':
                GoalPickup.objects.create(goal_id=goal_id)
            elif event == 'release':
                GoalPickup.objects.filter(goal_id=goal_id).delete()

        logger.info('Pickup monitor thread exiting')

    def pickup(self, goal_id):
        self.event_queue.put(('pickup', goal_id))

    def release(self, goal_id):
        self.event_queue.put(('release', goal_id))

    def shutdown(self):
        self.event_queue.shutdown()


class Middleware:
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __call__(self, goal, now, pickup_monitor=None):
        from .models import GoalState, _mark_as_failed

        # is it a killer task?
        GOALS_MAX_PICKUPS = getattr(settings, 'GOALS_MAX_PICKUPS', None)
        if (
            GOALS_MAX_PICKUPS is not None and
            goal.pickups.count() >= GOALS_MAX_PICKUPS
        ):
            logger.warning('Goal %s is a killer task, not pursuing it', goal.id)
            _mark_as_failed([goal.id], target_state=GoalState.IT_IS_A_KILLER_TASK)
            return None

        if pickup_monitor is not None:
            pickup_monitor.pickup(goal.id)

        progress = self.wrapped(goal, now, pickup_monitor=pickup_monitor)

        if pickup_monitor is not None:
            pickup_monitor.release(goal.id)

        return progress



import logging
import time

from django.utils import timezone

from django_goals.models import (
    handle_unblocked_goals, handle_waiting_for_date,
    handle_waiting_for_failed_preconditions, handle_waiting_for_preconditions,
    handle_waiting_for_worker, remove_old_goals,
)


logger = logging.getLogger(__name__)


def worker(stop_event=None, max_progress_count=float('inf'), once=False):
    """
    Worker is a busy-wait function that will keep checking for goals to pursue.
    It will keep running until stop_event is set.
    Process:
    1. Check if there are goals that are waiting for date and the date has come.
    2. Check if there are goals that are waiting for preconditions and all preconditions are achieved.
    3. Check if there are goals that are waiting for worker and pick one to pursue.
    4. If nothing could be done, sleep for a bit.
    5. Repeat until stop_event is set.
    """
    logger.info('Busy-wait worker started')
    progress_count = 0
    while (
        stop_event is None or
        not stop_event.is_set()
    ):
        if progress_count >= max_progress_count:
            logger.info('Max transitions reached, exiting')
            break

        transitions_done, local_progress_count = worker_turn(
            stop_event=stop_event,
            max_progress_count=max_progress_count - progress_count,
        )
        progress_count += local_progress_count

        if transitions_done == 0 and local_progress_count == 0:
            if once:
                logger.info('Nothing to do, exiting because of `once` flag')
                break
            # nothing could be done, let's go to sleep
            logger.debug('Nothing to do, sleeping for a bit')
            time.sleep(1)

    logger.info('Busy-wait worker exiting')


def worker_turn(now=None, stop_event=None, max_progress_count=float('inf')):
    """
    Worker turn is a single iteration of the worker.
    It will try to transition as many goals as possible.
    Returns a number of transitions done (all state changes)
    and a number of progress transitions done (real handler calls).
    """
    if now is None:
        now = timezone.now()
    transitions_done = 0
    transitions_done += handle_waiting_for_date(now)
    transitions_done += handle_waiting_for_preconditions()
    transitions_done += handle_waiting_for_failed_preconditions()
    transitions_done += handle_unblocked_goals()
    progress_count = 0
    while (
        stop_event is None or
        not stop_event.is_set()
    ):
        if progress_count >= max_progress_count:
            break
        did_a_thing = handle_waiting_for_worker()
        if not did_a_thing:
            break
        transitions_done += 1
        progress_count += 1
    remove_old_goals(now)
    return transitions_done, progress_count



# Generated by Django 5.1.4 on 2025-01-10 21:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('django_goals', '0007_goal_goals_achieved_idx'),
    ]

    operations = [
        migrations.RemoveIndex(
            model_name='goal',
            name='goals_waiting_for_precond_idx',
        ),
        migrations.AddField(
            model_name='goal',
            name='preconditions_mode',
            field=models.CharField(choices=[('all', 'All preconditions must be achieved before the goal can be pursued.'), ('any', 'Goal can be pursued if any of the preconditions is achieved.')], default='all', max_length=3),
        ),
        migrations.AddIndex(
            model_name='goal',
            index=models.Index(models.Q(('waiting_for_count__lte', 0)), condition=models.Q(('state', 'waiting_for_preconditions')), name='goals_waiting_for_precond_idx'),
        ),
        migrations.AddConstraint(
            model_name='goal',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(('preconditions_mode', 'any'), ('waiting_for_count__lte', 1)),
                    ('preconditions_mode', 'all'),
                    _connector='OR',
                ),
                name='goals_waiting_for_count_any',
            ),
        ),
    ]



import time
from unittest import mock

import pytest
from django.db import models
from django.test import override_settings
from django.utils import timezone

from django_goals.busy_worker import worker, worker_turn
from django_goals.models import (
    AllDone, Goal, GoalState, PreconditionFailureBehavior, schedule,
    thread_local,
)
from example_app.models import GoalRelatedModel

from .factories import GoalFactory, GoalProgressFactory


@pytest.mark.django_db
def test_worker_turn_noop():
    now = timezone.now()
    transitions_done = worker_turn(now)
    assert transitions_done == (0, 0)


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('failure_mode', 'expected_state', 'progress_count'),
    [
        (PreconditionFailureBehavior.PROCEED, GoalState.ACHIEVED, 1),
        (PreconditionFailureBehavior.BLOCK, GoalState.NOT_GOING_TO_HAPPEN_SOON, 0),
    ],
)
@override_settings(GOALS_GIVE_UP_AT=1)
def test_proceed_at_failure_mode_do_work(failure_mode, expected_state, progress_count):
    precond = schedule(fail)
    goal = schedule(
        noop,
        precondition_goals=[precond],
        precondition_failure_behavior=failure_mode,
    )
    worker(once=True)
    precond.refresh_from_db()
    assert precond.state == GoalState.GIVEN_UP
    goal.refresh_from_db()
    assert goal.state == expected_state
    assert goal.progress.count() == progress_count


def noop(goal):  # pylint: disable=unused-argument
    return AllDone()


def fail(goal):  # pylint: disable=unused-argument
    raise Exception('I failed!')


def trigger_database_error(goal):
    Goal.objects.create(id=goal.id)  # violates unique constraint


@pytest.mark.django_db(transaction=True)
def test_transaction_error_in_goal():
    goal = schedule(trigger_database_error)
    trasitions_count, progress_count = worker_turn(timezone.now())
    assert trasitions_count == 1
    assert progress_count == 1  # we called a handler and we must report it
    goal.refresh_from_db()
    assert goal.state == GoalState.WAITING_FOR_DATE
    # progress record is created
    assert goal.progress.exists()
    assert thread_local.current_goal is None


def use_lots_of_memory(goal):  # pylint: disable=unused-argument
    _unused = b'x' * 1024 * 1024 * 128  # 128 MiB  # noqa
    return AllDone()


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('memory_limit', 'expected_success'),
    [
        (None, True),
        (1, False),
        (128, False),
        (256, True),
    ],
)
def test_memory_limit(settings, memory_limit, expected_success):
    settings.GOALS_MEMORY_LIMIT_MIB = memory_limit
    # simulate we have some memory allocated outside of the goal handler
    _unused = b'x' * 1024 * 1024 * 2  # 2 MiB  # noqa
    goal = schedule(use_lots_of_memory)
    worker_turn(timezone.now())
    goal.refresh_from_db()
    expected_state = GoalState.ACHIEVED if expected_success else GoalState.WAITING_FOR_DATE
    assert goal.state == expected_state
    progress = goal.progress.get()
    assert progress.success == expected_success


def take_too_long(goal):  # pylint: disable=unused-argument
    time.sleep(2)
    return AllDone()


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('time_limit', 'expected_success'),
    [
        (None, True),
        (1, False),
        (3, True),
    ],
)
def test_time_limit(settings, time_limit, expected_success):
    settings.GOALS_TIME_LIMIT_SECONDS = time_limit
    goal = schedule(take_too_long)
    worker_turn(timezone.now())
    goal.refresh_from_db()
    expected_state = GoalState.ACHIEVED if expected_success else GoalState.WAITING_FOR_DATE
    assert goal.state == expected_state
    progress = goal.progress.get()
    assert progress.success == expected_success


@pytest.mark.django_db
@pytest.mark.parametrize(
    ('days_ago', 'state', 'expect_deleted'),
    [
        (31, GoalState.ACHIEVED, True),
        (1, GoalState.ACHIEVED, False),
        (31, GoalState.WAITING_FOR_WORKER, False),
        (31, GoalState.GIVEN_UP, False),
    ],
)
def test_old_achieved_goal_is_deleted(days_ago, state, expect_deleted):
    now = timezone.now()
    goal = GoalFactory(
        state=state,
        created_at=now - timezone.timedelta(days=days_ago),
    )
    GoalProgressFactory(goal=goal)
    dependent_goal = GoalFactory(precondition_goals=[goal])

    worker_turn(now)

    exists_after = Goal.objects.filter(id=goal.id).exists()
    assert exists_after is not expect_deleted

    # dependecy is removed
    dependent_goal.refresh_from_db()
    assert dependent_goal.precondition_goals.exists() is not expect_deleted


@pytest.mark.django_db
def test_protected_old_achieved_goal():
    now = timezone.now()
    goal = GoalFactory(
        state=GoalState.ACHIEVED,
        created_at=now - timezone.timedelta(days=31),
    )
    GoalRelatedModel.objects.create(goal=goal)

    # check we can't delete the goal
    with pytest.raises(models.ProtectedError):
        goal.delete()

    # worker turn doesn't crash, but emits a warning
    with mock.patch('django_goals.models.logger.warning') as warning:
        worker_turn(now)
    assert warning.call_count == 1
    assert 'old goals' in warning.call_args[0][0]
    assert 'protected' in str(warning.call_args[0][1])


def schedule_another(goal):
    schedule(schedule_another, blocked=True)
    return AllDone()


@pytest.mark.django_db
def test_deadline_is_inherited():
    now = timezone.now()
    goal = schedule(schedule_another, deadline=now + timezone.timedelta(days=1))
    worker_turn(now)
    another_goal = Goal.objects.exclude(id=goal.id).get()
    assert another_goal.deadline == goal.deadline



import logging
import re
import threading
import time
from contextlib import contextmanager
from datetime import timedelta

from django.core.management.base import BaseCommand

from django_goals.models import (
    handle_unblocked_goals, handle_waiting_for_date,
    handle_waiting_for_failed_preconditions, handle_waiting_for_preconditions,
    handle_waiting_for_worker, remove_old_goals,
)
from django_goals.pickups import PickupMonitorThread

from .goals_busy_worker import stop_signal_handler


logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = 'Run the worker'

    def add_arguments(self, parser):
        parser.add_argument(
            '--threads',
            action='append',  # This allows multiple instances of --threads
            help='Number of threads and optional deadline horizon (e.g., "3", "2:30m")',
        )
        parser.add_argument(
            '--once',
            action='store_true',
            help='Exit when no work is available',
        )

    def handle(self, *args, **options):
        # Parse the threads parameter
        threads_specs = options.get('threads') or ['1']  # Default to 1 worker if not specified
        worker_specs = []
        for spec in threads_specs:
            try:
                if ':' in spec:
                    # Format is N:HORIZON
                    count, horizon_str = spec.split(':', 1)
                    count = int(count)
                    horizon = parse_duration(horizon_str)
                else:
                    # Format is just N
                    count = int(spec)
                    horizon = None

                if count <= 0:
                    raise ValueError("Thread count must be a positive integer")
                worker_specs.append((count, horizon))

            except ValueError as e:
                self.stderr.write(f"Error parsing thread count '{spec}': {e}")
                return

        with stop_signal_handler() as stop_event:
            threaded_worker(
                worker_specs=worker_specs,
                stop_event=stop_event,
                once=options['once'],
            )


def parse_duration(duration_str):
    """
    Parse duration strings like "30m", "2h", "1d" into timedelta objects.
    Returns None if the string is "none", empty, or None.
    """
    if not duration_str or duration_str.lower() == 'none':
        return None

    units = {
        's': 'seconds',
        'm': 'minutes',
        'h': 'hours',
        'd': 'days',
        'w': 'weeks'
    }

    matches = re.match(r'^(\d+)([smhdw])$', duration_str.lower())
    if not matches:
        raise ValueError(f"Invalid duration format: {duration_str}. Use format like '30m', '2h', '1d'")

    value, unit = matches.groups()
    kwargs = {units[unit]: int(value)}
    return timedelta(**kwargs)


def threaded_worker(worker_specs=None, stop_event=None, once=False):
    if stop_event is None:
        stop_event = threading.Event()

    pickup_monitor = PickupMonitorThread()
    pickup_monitor.start()

    total_workers = sum(count for count, _ in worker_specs)
    workers_state = WorkersState(total_workers + 1)  # +1 for transitions thread

    threads = [
        TransitionsThread(
            stop_event=stop_event,
            once=once,
            workers_state=workers_state,
            thread_id="transitions",
        ),
    ]

    worker_id = 0
    for count, horizon in worker_specs:
        for i in range(count):
            threads.append(
                HeavyLiftingThread(
                    stop_event=stop_event,
                    once=once,
                    workers_state=workers_state,
                    thread_id=f"worker_{worker_id}",
                    deadline_horizon=horizon,
                    pickup_monitor=pickup_monitor,
                )
            )
            worker_id += 1

    for thread in threads:
        thread.start()

    for thread in threads:
        thread.join()

    pickup_monitor.shutdown()
    pickup_monitor.join()


class HeavyLiftingThread(threading.Thread):
    def __init__(
        self, stop_event, once, workers_state, thread_id, deadline_horizon=None,
        pickup_monitor=None,
    ):
        super().__init__()
        self.stop_event = stop_event
        self.once = once
        self.workers_state = workers_state
        self.thread_id = thread_id
        self.deadline_horizon = deadline_horizon
        self.pickup_monitor = pickup_monitor

    def run(self):
        logger.info('Busy-wait worker started, deadline_horizon: %s', self.deadline_horizon)

        while not self.stop_event.is_set():
            with self.workers_state.work_session(self.thread_id):
                try:
                    did_work = handle_waiting_for_worker(
                        deadline_horizon=self.deadline_horizon,
                        pickup_monitor=self.pickup_monitor,
                    )
                except Exception as e:
                    logger.exception(e)
                    # Treat exceptions as if we didn't do work
                    did_work = False

                self.workers_state.report_work(self.thread_id, did_work)

            if self.workers_state.all_idle and self.once:
                logger.info('All threads are idle. Exiting because of `once` flag.')
                break

            if not did_work:
                logger.debug('Nothing to do, sleeping for a bit')
                time.sleep(1)

        logger.info('Busy-wait worker exiting')


class TransitionsThread(threading.Thread):
    def __init__(self, stop_event, once, workers_state, thread_id):
        super().__init__()
        self.stop_event = stop_event
        self.once = once
        self.workers_state = workers_state
        self.thread_id = thread_id

    def run(self):
        logger.info('Transitions worker started')

        while not self.stop_event.is_set():
            with self.workers_state.work_session(self.thread_id):
                did_work = self._run_handlers()
                self.workers_state.report_work(self.thread_id, did_work)

            if self.workers_state.all_idle and self.once:
                logger.info('All threads are idle. Exiting because of `once` flag.')
                break

            if not did_work:
                logger.debug('Nothing to do, sleeping for a bit')
                time.sleep(1)

        logger.info('Transitions worker exiting')

    def _run_handlers(self):
        """Run all handlers and return True if any work was done"""
        try:
            results = [
                handle_waiting_for_preconditions(),
                handle_waiting_for_failed_preconditions(),
                handle_waiting_for_date(),
                handle_unblocked_goals(),
                remove_old_goals()
            ]
            return any(r for r in results if r)
        except Exception as e:
            logger.exception(e)
            return False  # No work done if there was an exception


class WorkersState:
    def __init__(self, thread_count):
        self.lock = threading.Lock()
        self.idle_threads = set()  # Threads that are permanently idle (no more work)
        self.active_threads = set()  # Threads currently registered as active
        self.total_threads = thread_count

    @contextmanager
    def work_session(self, thread_id):
        """
        Context manager to track a thread's work session.
        A thread enters a work session to check for and perform work.
        """
        with self.lock:
            self.active_threads.add(thread_id)

        yield

        with self.lock:
            assert thread_id not in self.active_threads, "You must call report_work before exiting the session"

    def report_work(self, thread_id, did_work):
        with self.lock:
            assert thread_id in self.active_threads, "Thread must be in work session to report work"
            self.active_threads.discard(thread_id)

            if did_work:
                # reactivate all, because the work we did might have unblocked other threads
                self.idle_threads.clear()

            else:
                # This thread is permanently idle
                self.idle_threads.add(thread_id)

    @property
    def all_idle(self):
        with self.lock:
            return len(self.idle_threads) == self.total_threads



import gc
import resource
import signal
from contextlib import contextmanager

from django.conf import settings


@contextmanager
def limit_memory():
    limit_mib = getattr(settings, 'GOALS_MEMORY_LIMIT_MIB', None)
    if limit_mib is None:
        yield
        return
    original_limit_soft, original_limit_hard = resource.getrlimit(resource.RLIMIT_AS)
    try:
        resource.setrlimit(resource.RLIMIT_AS, (limit_mib * 1024 * 1024, original_limit_hard))
        yield
    except MemoryError:
        gc.collect()
        raise
    finally:
        resource.setrlimit(resource.RLIMIT_AS, (original_limit_soft, original_limit_hard))


class TimesUp(Exception):
    pass


def sigalrm_handler(signum, frame):
    raise TimesUp()


@contextmanager
def limit_time():
    seconds = getattr(settings, 'GOALS_TIME_LIMIT_SECONDS', None)
    if seconds is None:
        yield
        return
    previous_handler = signal.signal(signal.SIGALRM, sigalrm_handler)
    assert previous_handler in (signal.SIG_DFL, sigalrm_handler)
    try:
        signal.alarm(seconds)
        yield
    finally:
        signal.alarm(0)



# Generated by Django 5.2 on 2025-06-04 21:11

import uuid

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("django_goals", "0010_remove_goal_goals_waiting_for_failed_idx_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="GoalPickup",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("created_at", models.DateTimeField(default=django.utils.timezone.now)),
                (
                    "goal",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="pickups",
                        to="django_goals.goal",
                    ),
                ),
            ],
            options={
                "ordering": ("-created_at",),
            },
        ),
    ]



import datetime
import inspect
import logging
import threading
import time
import uuid

from django.conf import settings
from django.db import connections, models, transaction
from django.db.models.functions import Least
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from .limits import limit_memory, limit_time
from .notifications import (
    listen_goal_progress, notify_goal_progress, notify_goal_waiting_for_worker,
)
from .pickups import GoalPickup


__all__ = (
    'GoalPickup',
)


logger = logging.getLogger(__name__)


class GoalState(models.TextChoices):
    # Goal is explicitly marked not to be pursued
    BLOCKED = 'blocked'
    # Goal cannot be pursued yet, because it is allowed only after future date
    WAITING_FOR_DATE = 'waiting_for_date'
    # Goal cannot be pursued yet, because other goals need to be achieved first
    WAITING_FOR_PRECONDITIONS = 'waiting_for_preconditions'
    # Goal is ready to be pursued. We are waiting for a worker to pick it up
    WAITING_FOR_WORKER = 'waiting_for_worker'
    # The goal has been achieved
    ACHIEVED = 'achieved'
    # Too many failed attempts when pursuing the goal
    GIVEN_UP = 'given_up'
    # Goal is waiting on a precondition that wont be achieved
    NOT_GOING_TO_HAPPEN_SOON = 'not_going_to_happen_soon'
    # Many workers perished without a trace working on this goal
    IT_IS_A_KILLER_TASK = 'it_is_a_killer_task'


NOT_GOING_TO_HAPPEN_SOON_STATES = (
    GoalState.BLOCKED,
    GoalState.GIVEN_UP,
    GoalState.NOT_GOING_TO_HAPPEN_SOON,
    GoalState.IT_IS_A_KILLER_TASK,
)


WAITING_STATES = (
    GoalState.WAITING_FOR_DATE,
    GoalState.WAITING_FOR_PRECONDITIONS,
    GoalState.WAITING_FOR_WORKER,
)


class PreconditionsMode(models.TextChoices):
    ALL = 'all', _('All preconditions must be achieved before the goal can be pursued.')
    ANY = 'any', _('Goal can be pursued if any of the preconditions is achieved.')


class PreconditionFailureBehavior(models.TextChoices):
    BLOCK = 'block', _('Do not proceed if preconditions fail')
    PROCEED = 'proceed', _('Proceed with goal execution even if preconditions fail')


class Goal(models.Model):
    """
    Goal represents a state we want to achieve.
    Goal will be pursued by calling a handler function.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    state = models.CharField(
        max_length=30,
        db_index=True,
        choices=GoalState.choices,
        default=GoalState.WAITING_FOR_DATE,
    )
    handler = models.CharField(max_length=100)
    instructions = models.JSONField(null=True)
    precondition_date = models.DateTimeField(
        default=timezone.now,
        help_text=_('Goal will not be pursued before this date.'),
    )

    precondition_goals = models.ManyToManyField(
        to='self',
        symmetrical=False,
        related_name='dependent_goals',
        through='GoalDependency',
        blank=True,
    )
    preconditions_mode = models.CharField(
        max_length=3,
        choices=PreconditionsMode.choices,
        default=PreconditionsMode.ALL,
    )
    precondition_failure_behavior = models.CharField(
        max_length=10,
        choices=PreconditionFailureBehavior.choices,
        default=PreconditionFailureBehavior.BLOCK,
    )
    waiting_for_count = models.IntegerField(
        default=0,
        help_text=_('Number of precondition goals that must finish before this goal can be pursued.'),
    )
    waiting_for_not_achieved_count = models.IntegerField(  # for ALL mode this is the same as waiting_for_count
        default=0,
        help_text=_('Number of precondition goals that are not achieved yet.'),
    )
    waiting_for_failed_count = models.IntegerField(
        default=0,
        help_text=_('Number of precondition goals that failed.'),
    )

    deadline = models.DateTimeField(
        default=timezone.now,
        help_text=_('Goals having deadline sooner will be pursued first.'),
    )
    created_at = models.DateTimeField(default=timezone.now, db_index=True)

    class Meta:
        ordering = ('-created_at',)
        indexes = [
            models.Index(
                fields=['precondition_date'],
                condition=models.Q(state=GoalState.WAITING_FOR_DATE),
                name='goals_waiting_for_date_idx',
            ),
            models.Index(
                models.Q(waiting_for_count__lte=0),
                condition=models.Q(state=GoalState.WAITING_FOR_PRECONDITIONS),
                name='goals_waiting_for_precond_idx',
            ),
            models.Index(
                fields=['deadline'],
                condition=models.Q(state=GoalState.WAITING_FOR_WORKER),
                name='goals_waiting_for_worker_idx',
            ),
            models.Index(  # for blocking goals that are waiting for blocked preconds
                fields=['waiting_for_failed_count'],
                condition=models.Q(
                    state=GoalState.WAITING_FOR_PRECONDITIONS,
                    precondition_failure_behavior=PreconditionFailureBehavior.BLOCK,
                ),
                name='goals_waiting_for_failed_idx',
            ),
            models.Index(  # for unblocking goals when preconditions becomes unblocked
                fields=['waiting_for_failed_count'],
                condition=models.Q(state=GoalState.NOT_GOING_TO_HAPPEN_SOON),
                name='goals_unblocking_idx',
            ),
            models.Index(  # for deleting old done goals
                fields=['created_at'],
                condition=models.Q(state=GoalState.ACHIEVED),
                name='goals_achieved_idx',
            ),
        ]
        constraints = [
            models.CheckConstraint(
                check=models.Q(
                    waiting_for_count__lte=1,
                    preconditions_mode=PreconditionsMode.ANY,
                ) | models.Q(
                    preconditions_mode=PreconditionsMode.ALL,
                ),
                name='goals_waiting_for_count_any',
            ),
            models.CheckConstraint(
                check=models.Q(precondition_failure_behavior__in=PreconditionFailureBehavior.values),
                name='goals_precondition_failure_behavior',
            ),
        ]

    def __str__(self):
        return f'{self.handler} ({self.state})'


class GoalDependency(models.Model):
    """
    The GoalDependency is a many-to-many through model for Goal.
    Its purpose is to define which goals need to be achieved before another goal can be pursued.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    dependent_goal = models.ForeignKey(
        to=Goal,
        on_delete=models.CASCADE,
        related_name='dependencies',
    )
    precondition_goal = models.ForeignKey(
        to=Goal,
        on_delete=models.PROTECT,
        related_name='dependents',
    )

    class Meta:
        unique_together = (
            ('dependent_goal', 'precondition_goal'),
        )


class GoalProgress(models.Model):
    """
    GoalProgress represents a single attempt to achieve a goal.
    """
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    goal = models.ForeignKey(Goal, on_delete=models.CASCADE, related_name='progress')
    success = models.BooleanField()
    created_at = models.DateTimeField(default=timezone.now)
    time_taken = models.DurationField(null=True)
    message = models.CharField(max_length=200, blank=True)

    class Meta:
        ordering = ('goal', '-created_at')


@transaction.atomic
def block_goal(goal_id):
    """
    Mark the goal as blocked, so it will not be pursued.
    """
    goal = Goal.objects.select_for_update().get(id=goal_id)
    if goal.state not in WAITING_STATES:
        raise ValueError(f'Cannot block goal in state {goal.state}')
    _mark_as_failed([goal_id], target_state=GoalState.BLOCKED)
    return goal


@transaction.atomic
def unblock_retry_goal(goal_id):
    """
    Mark the goal as unblocked, so it can be pursued again.
    """
    goal = Goal.objects.select_for_update().get(id=goal_id)
    if goal.state not in NOT_GOING_TO_HAPPEN_SOON_STATES:
        raise ValueError(f'Cannot unblock/retry goal in state {goal.state}')
    _mark_as_unfailed([goal_id])
    GoalPickup.objects.filter(goal_id=goal_id).delete()
    return goal


def _mark_as_failed(goal_ids, target_state):
    """
    All goal must be in waititng state.
    You must be in a transaction and have a lock on the goals.
    """
    assert target_state in NOT_GOING_TO_HAPPEN_SOON_STATES
    if not goal_ids:
        return
    Goal.objects.filter(id__in=goal_ids).update(state=target_state)
    # update waiting-for failed count in dependent goals
    Goal.objects.filter(
        precondition_goals__id__in=goal_ids,
    ).update(
        waiting_for_failed_count=models.F('waiting_for_failed_count') + 1,
    )
    Goal.objects.filter(
        precondition_goals__id__in=goal_ids,
        precondition_failure_behavior=PreconditionFailureBehavior.PROCEED,
    ).update(
        waiting_for_count=models.F('waiting_for_count') - 1,
    )


def _mark_as_unfailed(goal_ids):
    """
    All goal_ids must be in NOT_GOING_TO_HAPPEN_SOON_STATES.
    You must be in a transaction and have a lock on the goals.
    """
    if not goal_ids:
        return
    Goal.objects.filter(id__in=goal_ids).update(state=GoalState.WAITING_FOR_DATE)
    # update waiting-for failed count in dependent goals
    Goal.objects.filter(
        precondition_goals__id__in=goal_ids,
    ).update(
        waiting_for_failed_count=models.F('waiting_for_failed_count') - 1,
    )


@transaction.atomic
def handle_waiting_for_date(now=None):
    """
    Transition goals that are waiting for precondition date and the date has come.
    """
    if now is None:
        now = timezone.now()
    qs = Goal.objects.filter(
        state=GoalState.WAITING_FOR_DATE,
        precondition_date__lte=now,
    ).select_for_update(
        skip_locked=True,
        no_key=True,
    )
    ids = list(qs.values_list('id', flat=True))
    return Goal.objects.filter(
        id__in=ids,
    ).update(state=GoalState.WAITING_FOR_PRECONDITIONS)


@transaction.atomic
def handle_waiting_for_preconditions(goals_qs=None):
    """
    Transition goals that are waiting for precondition goals to be achieved
    and all (or any, depending on mode) preconditions are achieved.
    """
    if goals_qs is None:
        goals_qs = Goal.objects.all()
    transitions_done = 0

    new_waiting_for_worker = goals_qs.filter(
        state=GoalState.WAITING_FOR_PRECONDITIONS,
        waiting_for_count__lte=0,
    ).select_for_update(
        no_key=True,
        skip_locked=True,
    ).values_list('id', flat=True)
    new_waiting_for_worker = list(new_waiting_for_worker)
    if new_waiting_for_worker:
        Goal.objects.filter(id__in=new_waiting_for_worker).update(state=GoalState.WAITING_FOR_WORKER)
        with connections['default'].cursor() as cursor:
            for goal_id in new_waiting_for_worker:
                notify_goal_waiting_for_worker(cursor, goal_id)
    transitions_done += len(new_waiting_for_worker)

    return transitions_done


@transaction.atomic
def handle_waiting_for_failed_preconditions():
    """
    if a goal is waiting for preconditions that are failed, it's not going to happen soon
    """
    goals_qs = Goal.objects.all()
    transitions_done = 0

    new_failed = goals_qs.filter(
        state=GoalState.WAITING_FOR_PRECONDITIONS,
        precondition_failure_behavior=PreconditionFailureBehavior.BLOCK,
        waiting_for_failed_count__gt=0,
    ).select_for_update(
        no_key=True,
        skip_locked=True,
    ).values_list('id', flat=True)
    new_failed = list(new_failed)
    _mark_as_failed(new_failed, target_state=GoalState.NOT_GOING_TO_HAPPEN_SOON)
    transitions_done += len(new_failed)

    return transitions_done


@transaction.atomic
def handle_unblocked_goals():
    """
    Transition goals that have no failed preconditions, yet are marked as not going to happen soon.
    """
    qs = Goal.objects.filter(
        state=GoalState.NOT_GOING_TO_HAPPEN_SOON,
        waiting_for_failed_count__lte=0,
    ).select_for_update(
        skip_locked=True,
        no_key=True,
    )
    ids = list(qs.values_list('id', flat=True))
    _mark_as_unfailed(ids)
    return len(ids)


@transaction.atomic
def handle_waiting_for_worker(deadline_horizon=None, pickup_monitor=None):
    """
    Transition goals that are waiting for a worker to pick them up.
    """
    now = timezone.now()
    # Get the first goal that is waiting for a worker
    qs = Goal.objects.filter(state=GoalState.WAITING_FOR_WORKER).order_by(
        'deadline',
    )
    if deadline_horizon is not None:
        qs = qs.filter(deadline__lte=now + deadline_horizon)
    goal = qs.select_for_update(
        skip_locked=True,
        no_key=True,
    ).first()
    if goal is None:
        # nothing to do
        return None

    return _pursue_goal(goal, now, pickup_monitor=pickup_monitor)


class FsckMiddleware:
    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __call__(self, goal, now, pickup_monitor=None):
        if now < goal.precondition_date:
            logger.warning('Precondition date bug in goal %s. Precondition date is in the future', goal.id)
        if (
            goal.preconditions_mode == PreconditionsMode.ALL and
            goal.waiting_for_count != 0
        ):
            logger.warning('Waiting-for bug in goal %s. Expected 0, got %s', goal.id, goal.waiting_for_count)
        if (
            goal.preconditions_mode == PreconditionsMode.ANY and
            goal.waiting_for_count > 0
        ):
            logger.warning('Waiting-for (ANY mode) bug in goal %s. Expected 0 or less, got %s', goal.id, goal.waiting_for_count)
        if goal.waiting_for_failed_count < 0:
            logger.warning('Waiting-for-failed bug in goal %s. Expected 0 or more, got %s', goal.id, goal.waiting_for_failed_count)
        if (
            goal.precondition_failure_behavior == PreconditionFailureBehavior.BLOCK and
            goal.waiting_for_failed_count != 0
        ):
            logger.warning('Waiting-for-failed (BLOCK mode) bug in goal %s. Expected 0, got %s', goal.id, goal.waiting_for_failed_count)

        return self.wrapped(goal, now, pickup_monitor=pickup_monitor)


def _pursue_goal_core(goal, now, pickup_monitor=None):
    logger.info('Just about to pursue goal %s: %s', goal.id, goal.handler)
    start_time = time.monotonic()
    try:
        ret = follow_instructions(goal)

    except Exception:  # pylint: disable=broad-except
        logger.exception('Goal %s failed', goal.id)
        success = False
        message = ''
        failure_index = goal.progress.filter(success=False).count()
        retry_delay = get_retry_delay(failure_index)
        if retry_delay is None:
            goal.state = GoalState.GIVEN_UP
        else:
            goal.state = GoalState.WAITING_FOR_DATE
            goal.precondition_date = now + retry_delay

    else:
        if isinstance(ret, RetryMeLater):
            logger.info('Goal %s needs to be retried later. Message: %s', goal.id, ret.message)
            success = True
            message = ret.message
            goal.state = GoalState.WAITING_FOR_DATE
            if ret.precondition_date is not None:
                goal.precondition_date = max(goal.precondition_date, ret.precondition_date)
            _add_precondition_goals(goal, ret.precondition_goals)

        elif isinstance(ret, AllDone):
            logger.info('Goal %s was achieved', goal.id)
            success = True
            message = ''
            goal.state = GoalState.ACHIEVED

        else:
            logger.warning('Goal %s handler returned unknown value, which is ignored', goal.id)
            success = True
            message = ''
            goal.state = GoalState.ACHIEVED

    time_taken = time.monotonic() - start_time

    # decrease waiting-for counter in dependent goals
    if goal.state == GoalState.ACHIEVED:
        Goal.objects.filter(
            precondition_goals=goal,
        ).update(
            waiting_for_count=models.F('waiting_for_count') - 1,
            waiting_for_not_achieved_count=models.F('waiting_for_not_achieved_count') - 1,
        )

    progress = goal.progress.create(
        success=success,
        created_at=now,
        time_taken=datetime.timedelta(seconds=time_taken),
        message=message,
    )

    # check max progress count
    max_progress_count = getattr(settings, 'GOALS_MAX_PROGRESS_COUNT', 100)
    if (
        max_progress_count is not None and
        goal.state != GoalState.ACHIEVED and
        goal.progress.count() >= max_progress_count
    ):
        logger.warning('Goal %s reached max progress count, giving up', goal.id)
        goal.state = GoalState.GIVEN_UP

    # incrase waiting-for failed count in dependent goals
    if goal.state == GoalState.GIVEN_UP:
        # goal.state will be saved twice, but it's fine
        _mark_as_failed([goal.id], target_state=GoalState.GIVEN_UP)

    goal.save(update_fields=['state', 'precondition_date'])
    notify_goal_progress(goal.id, goal.state)

    return progress


_pursue_goal = _pursue_goal_core
for middleware_path in reversed(getattr(settings, 'GOALS_MIDDLEWARE', [
    'django_goals.pickups.Middleware',
    'django_goals.models.FsckMiddleware',
])):
    middleware = import_string(middleware_path)
    _pursue_goal = middleware(_pursue_goal)


class GoalsThreadLocal(threading.local):
    def __init__(self):
        self.current_goal = None


thread_local = GoalsThreadLocal()


@limit_time()
@limit_memory()
# This is a savepoint that protects worker transaction so that we can always report progress.
# Raw error in trnsaction would prevent from executing any queries inside it.
@transaction.atomic
def follow_instructions(goal):
    """
    Call the handler function with instructions.
    """
    func = import_string(goal.handler)
    instructions = goal.instructions
    if instructions is None:
        instructions = {}
    assert thread_local.current_goal is None
    thread_local.current_goal = goal
    try:
        return func(
            goal,
            *instructions.get('args', ()),
            **instructions.get('kwargs', {}),
        )
    except RetryMeLaterException as e:
        return RetryMeLater(**e.__dict__)
    finally:
        thread_local.current_goal = None


def get_retry_delay(failure_index):
    """
    Get the delay before retrying the goal.
    failure_index is how many times the goal has failed (before this time). So first time this is called with 0.
    """
    current_failure_index = failure_index + 1
    give_up_at = getattr(settings, 'GOALS_GIVE_UP_AT', 4)
    if current_failure_index >= give_up_at:
        return None
    return datetime.timedelta(seconds=10) * (2 ** failure_index)


def remove_old_goals(now=None):
    if now is None:
        now = timezone.now()
    retention_seconds = getattr(settings, 'GOALS_RETENTION_SECONDS', 60 * 60 * 24 * 7)
    if retention_seconds is None:
        return 0
    try:
        with transaction.atomic():
            ids_to_delete = Goal.objects.filter(
                state=GoalState.ACHIEVED,
                created_at__lt=now - datetime.timedelta(seconds=retention_seconds),
            ).select_for_update(
                skip_locked=True,
            ).values_list('id', flat=True)
            ids_to_delete = list(ids_to_delete[:100])
            if not ids_to_delete:
                return 0
            GoalDependency.objects.filter(precondition_goal_id__in=ids_to_delete).delete()
            GoalDependency.objects.filter(dependent_goal_id__in=ids_to_delete).delete()
            Goal.objects.filter(id__in=ids_to_delete).delete()
            logger.info('Deleted %s old, achieved goals', len(ids_to_delete))
            return len(ids_to_delete)
    except models.ProtectedError as e:
        logger.warning('When cleaning old goals: %s', e)
        return 0


class RetryMeLater:
    """
    Like a process yielding in operating system.
    """
    def __init__(self, precondition_date=None, precondition_goals=(), message=''):
        self.precondition_date = precondition_date
        self.precondition_goals = precondition_goals
        self.message = message


class RetryMeLaterException(RetryMeLater, Exception):
    pass


class AllDone:
    pass


def schedule(
    func, args=None, kwargs=None,
    precondition_date=None, precondition_goals=None, blocked=False,
    deadline=None,
    listen=False,
    preconditions_mode=PreconditionsMode.ALL,
    precondition_failure_behavior=PreconditionFailureBehavior.BLOCK,
):
    """
    Schedule a goal to be pursued.
    """
    state = GoalState.WAITING_FOR_DATE

    instructions = {}
    if args is not None:
        instructions['args'] = args
    if kwargs is not None:
        instructions['kwargs'] = kwargs
    if not instructions:
        instructions = None

    if precondition_date is None:
        precondition_date = timezone.now()
        state = GoalState.WAITING_FOR_PRECONDITIONS
    if precondition_goals is None:
        precondition_goals = []
    if (
        not precondition_goals and
        state == GoalState.WAITING_FOR_PRECONDITIONS
    ):
        state = GoalState.WAITING_FOR_WORKER
    if blocked:
        state = GoalState.BLOCKED
    func_name = inspect.getmodule(func).__name__ + '.' + func.__name__

    if deadline is None and thread_local.current_goal is not None:
        deadline = thread_local.current_goal.deadline
    elif deadline is None:
        default_deadline_delta = datetime.timedelta(
            seconds=getattr(settings, 'GOALS_DEFAULT_DEADLINE_SECONDS', 7 * 24 * 60 * 60),
        )
        deadline = timezone.now() + default_deadline_delta

    goal = Goal(
        state=state,
        handler=func_name,
        instructions=instructions,
        precondition_date=precondition_date,
        deadline=deadline,
        preconditions_mode=preconditions_mode,
        precondition_failure_behavior=precondition_failure_behavior,
    )

    return _schedule(goal, precondition_goals, listen=listen)


def _schedule_core(goal, precondition_goals, listen):
    if listen:
        listen_goal_progress(goal.id)

    with transaction.atomic():
        goal.save()
        _add_precondition_goals(goal, precondition_goals)
        if goal.state == GoalState.WAITING_FOR_WORKER:
            with connections['default'].cursor() as cursor:
                notify_goal_waiting_for_worker(cursor, goal.id)

    return goal


_schedule = _schedule_core
for middleware_path in reversed(getattr(settings, 'GOALS_SCHEDULE_MIDDLEWARE', [])):
    middleware = import_string(middleware_path)
    _schedule = middleware(_schedule)


def _add_precondition_goals(goal, precondition_goals):
    # We can be sure current waiting count are zero, because goal can add preconditions only
    # in the handler or when scheduling a new goal.
    # However, waiting_for_not_achieved_count can be non-zero when running in ANY mode,
    # and waiting_for_failed_count can be non-zero when running in PROCEED precond failure mode.
    goal.waiting_for_count = 0

    if precondition_goals is None:
        goal.save(update_fields=[
            'waiting_for_count',
        ])
        return

    # Lock and retrieve new precondition goals that aren't already dependencies
    # This locking is critical to prevent a race condition where:
    # 1. We check a precondition goal's state (not achieved)
    # 2. Before we create the dependency, that precondition becomes achieved
    # 3. The achievement event can't decrease our waiting_for_count because the dependency doesn't exist yet
    # 4. We set waiting_for_count=1 based on step 1's observation
    # 5. Result: waiting_for_count stays at 1 forever, never reaching zero
    new_precondition_goals = list(Goal.objects.filter(
        id__in=[g.id for g in precondition_goals],
    ).exclude(
        dependent_goals=goal,
    ).select_for_update(no_key=True))

    # add to our preconditions
    goal.precondition_goals.add(*new_precondition_goals)

    # update waiting-for counters
    for precondition_goal in new_precondition_goals:
        if precondition_goal.state != GoalState.ACHIEVED:
            goal.waiting_for_count += 1
            goal.waiting_for_not_achieved_count += 1
        if precondition_goal.state in NOT_GOING_TO_HAPPEN_SOON_STATES:
            goal.waiting_for_failed_count += 1

    if goal.precondition_failure_behavior == PreconditionFailureBehavior.PROCEED:
        # in PROCEED precond mode, failed preconditions are treated like achieved
        goal.waiting_for_count -= goal.waiting_for_failed_count

    if goal.preconditions_mode == PreconditionsMode.ANY:
        # cap waiting_for_count at 1 in ANY mode
        goal.waiting_for_count = min(goal.waiting_for_count, 1)
        # ensure we are waiting for something if there are any not achieved preconditions
        if goal.waiting_for_not_achieved_count > 0:
            goal.waiting_for_count = 1
        # Detect the case where some precondition completed in the span between
        # handler checked the preconditions and we locked them.
        # We assume that precondition_goals contain the version checked by the handler.
        # This is releavnt only for ANY precond mode - because we are interested in act of
        # precond becoming achieved, not the final state like in ALL mode.
        orig_preconds_by_id = {g.id: g for g in precondition_goals}
        for precondition_goal in new_precondition_goals:
            orig_goal = orig_preconds_by_id[precondition_goal.id]
            if (
                orig_goal.state != GoalState.ACHIEVED and
                precondition_goal.state == GoalState.ACHIEVED
            ) or (
                goal.precondition_failure_behavior == PreconditionFailureBehavior.PROCEED and
                orig_goal.state not in NOT_GOING_TO_HAPPEN_SOON_STATES and
                precondition_goal.state in NOT_GOING_TO_HAPPEN_SOON_STATES
            ):
                goal.waiting_for_count = 0

    goal.save(update_fields=[
        'waiting_for_count',
        'waiting_for_not_achieved_count',
        'waiting_for_failed_count',
    ])

    # move deadline earlier for preconditions, if needed
    update_goals_deadline(Goal.objects.filter(
        id__in=[g.id for g in new_precondition_goals],
    ), goal.deadline)


def update_goals_deadline(goals_qs, deadline):
    goals_to_be_updated = list(goals_qs.filter(
        deadline__gt=deadline,
    ).exclude(
        state=GoalState.ACHIEVED,
    ))
    Goal.objects.filter(
        id__in=[goal.id for goal in goals_to_be_updated],
    ).update(deadline=Least('deadline', models.Value(deadline)))
    for goal in goals_to_be_updated:
        update_goals_deadline(goal.precondition_goals.all(), deadline)
